@@ -1,6 +1,6 @@
 package Mojolicious::Plugin::Authentication;
-BEGIN {
-  $Mojolicious::Plugin::Authentication::VERSION = '1.21';
+{
+  $Mojolicious::Plugin::Authentication::VERSION = '1.22';
 }
 use Mojo::Base 'Mojolicious::Plugin';
 
@@ -14,12 +14,19 @@ sub register {
     die __PACKAGE__, ": missing 'validate_user' subroutine ref in parameters\n"
         unless $args->{validate_user} && ref($args->{validate_user}) eq 'CODE';
 
-    my $lazy_mode         = $args->{lazy_mode}   || 0;
+    if(defined($args->{lazy})) {
+        warn __PACKAGE__, ': the "lazy" option is deprecated, use "autoload_user" instead', "\n";
+        $args->{autoload_user} = delete($args->{lazy});
+    }
+
+    my $autoload_user     = defined($args->{autoload_user}) ? $args->{autoload_user} : 0;
     my $session_key       = $args->{session_key} || 'auth_data';
     my $our_stash_key     = $args->{stash_key}   || '__authentication__';
     my $load_user_cb      = $args->{load_user};
     my $validate_user_cb  = $args->{validate_user};
+    my $current_user_fn   = $args->{current_user_fn} || 'current_user';
 
+    # Unconditionally load the user based on uid in session
     my $user_loader_sub = sub {
         my $c = shift;
 
@@ -28,24 +35,23 @@ sub register {
             if ($user) {
                 $c->stash($our_stash_key => { user => $user });
             }
-            elsif ($lazy_mode) {
+            else {
+                # cache result that user does not exist
                 $c->stash($our_stash_key => { no_user => 1 });
             }
         }
     };
 
+    # Fetch the current user object from the stash - loading it if not already loaded
     my $user_stash_extractor_sub = sub {
         my $c = shift;
-        my $is_return_binary = shift || 0;
 
-        if (
-            $lazy_mode
-            && !(
+        if ( !(
                 defined($c->stash($our_stash_key))
                 && ($c->stash($our_stash_key)->{no_user}
                     || defined($c->stash($our_stash_key)->{user}))
+              )
             )
-          )
         {
             $user_loader_sub->($c);
         }
@@ -53,19 +59,15 @@ sub register {
         my $user_def = defined($c->stash($our_stash_key))
                           && defined($c->stash($our_stash_key)->{user});
 
-        return $is_return_binary
-          ? ($user_def ? 1 : 0)
-          : ($user_def ? $c->stash($our_stash_key)->{user} : undef);
+        return $user_def ? $c->stash($our_stash_key)->{user} : undef;
 
     };
 
-    if (!$lazy_mode) {
-        $app->hook(before_dispatch => $user_loader_sub);
-    }
+    $app->hook(before_dispatch => $user_loader_sub) if($autoload_user);
 
     $app->routes->add_condition(authenticated => sub {
         my ($r, $c, $captures, $required) = @_;
-        return ($required && $c->user_exists) ? 1 : 0;
+        return ($required && $c->is_user_authenticated) ? 1 : 0;
     });
 
     $app->routes->add_condition(signed => sub {
@@ -73,20 +75,40 @@ sub register {
         return ($required && $c->signature_exists) ? 1 : 0;
     });
 
+    # deprecation handling
+    $app->helper(user_exists => sub {
+        warn __PACKAGE__, ': the "user_exists" helper is deprecated, use "is_user_authenticated" instead', "\n";
+        return shift->is_user_authenticated(@_);
+    });
+
+    $app->helper(user => sub {
+        warn __PACKAGE__, ': the "user" helper is deprecated, use "', $current_user_fn, '" instead', "\n";
+        return shift->$current_user_fn(@_);
+    });
+
+    my $current_user = sub {
+        my $c = shift;
+        return $user_stash_extractor_sub->($c);
+    };
+
+    $app->helper(reload_user => sub {
+        my $c = shift;
+        # Clear stash to force a reload of the user object
+        delete $c->stash->{$our_stash_key};
+        return $current_user->($c);
+    });
+
     $app->helper(signature_exists => sub {
         my $c = shift;
         return $c->session($session_key) ? 1 : 0;
     });
 
-    $app->helper(user_exists => sub {
+    $app->helper(is_user_authenticated => sub {
         my $c = shift;
-        return $user_stash_extractor_sub->($c, 1);
+        return defined($current_user->($c)) ? 1 : 0;
     });
 
-    $app->helper(user => sub {
-        my $c = shift;
-        return $user_stash_extractor_sub->($c);
-    });
+    $app->helper("$current_user_fn" => $current_user);
 
     $app->helper(logout => sub {
         my $c = shift;
@@ -98,8 +120,11 @@ sub register {
         my ($c, $user, $pass, $extradata) = @_;
         if (my $uid = $validate_user_cb->($c, $user, $pass, $extradata)) {
             $c->session($session_key => $uid);
-            $c->stash->{$our_stash_key}->{user} = $load_user_cb->($c, $uid);
-            return 1;
+            # Clear stash to force reload of any already loaded user object
+            delete $c->stash->{$our_stash_key};
+            return 1 if defined( $current_user->($c) );
+            # Failed to load user object. Perhaps some kind of race condition or other error?
+            return;
         }
         return;
     });
@@ -113,17 +138,18 @@ Mojolicious::Plugin::Authentication - A plugin to make authentication a bit easi
 
 =head1 VERSION
 
-version 1.21
+version 1.22
 
 =head1 SYNOPSIS
 
     use Mojolicious::Plugin::Authentication
 
     $self->plugin('authentication' => {
-        'lazy_mode' => 1,
+        'autoload_user' => 1,
         'session_key' => 'wickedapp',
         'load_user' => sub { ... },
         'validate_user' => sub { ... },
+        'current_user_fn' => 'user', # compatibility with old code
     });
 
     if ($self->authenticate('username', 'password', { optional => 'extra data stuff' })) {
@@ -137,14 +163,18 @@ version 1.21
 
 Authenticate will use the supplied C<load_user> and C<validate_user> subroutine refs to see whether a user exists with the given username and password, and will set up the session accordingly.  Returns true when the user has been successfully authenticated, false otherwise. You can pass additional data along in the extra_data hashref, it will be passed to your C<validate_user> subroutine as-is.
 
-=head2 user_exists
+=head2 is_user_authenticated
 
-Returns true if an authenticated user exists, false otherwise.
+Returns true if current_user() returns some valid object, false otherwise.
 
-=head2 user
+=head2 current_user
 
 Returns the user object as it was returned from the supplied C<load_user> subroutine ref.
 
+=head2 reload_user
+
+Flushes the current user object and then returns user().
+
 =head2 signature_exists
 
 Returns true if uid signature exist on the client side (in cookies), false otherwise.
@@ -167,7 +197,9 @@ The following options can be set for the plugin:
 
 =item session_key (optional) The name of the session key
 
-=item lazy_mode (optional) Turn on 'lazy mode' - user data to be loaded only if it be used. May reduce site latency in some cases.
+=item autoload_user (optional) Turn on/off automatic loading of user data - user data can be loaded only if it be used. May reduce site latency in some cases.
+
+=item current_user_fn (optional) Set the name for the current_user() helper function
 
 =back 
 
@@ -229,7 +261,7 @@ If you want to be able to send people to a login page, you will have to use the
     my $members_only = $r->route('/members')->to(cb => sub {
         my $self = shift;
 
-        $self->redirect_to('/login') and return 0 unless($self->user_exists);
+        $self->redirect_to('/login') and return 0 unless($self->is_user_authenticated);
         return 1;
     });
 
@@ -257,7 +289,7 @@ And in your Auth controller you would put:
 
     sub check {
         my $self = shift;
-        $self->redirect_to('/login') and return 0 unless($self->user_exists);
+        $self->redirect_to('/login') and return 0 unless($self->is_user_authenticated);
         return 1;
     });
 
@@ -324,6 +356,10 @@ Ivo Welch
     -   For donating his first ever Mojolicious application that shows an example of how
         to use this module
 
+Ed W
+    -   Adding the current_user() functionality, as well as some method renaming to make things
+        a bit more sane.
+
 =head1 LICENSE AND COPYRIGHT
 
 Copyright 2011-2012 Ben van Staveren.
@@ -335,4 +371,4 @@ by the Free Software Foundation; or the Artistic License.
 See http://dev.perl.org/licenses/ for more information.
 
 
-=cut
\ No newline at end of file
+=cut
